package org.geekfu.Cartographer;

import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.Serializable;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.border.Border;
import javax.swing.event.MouseInputListener;

import org.geekfu.Cartographer.MapCell;

/**
 * @author randrews
 * This class represents a tile-based map.
 * It is composed of a square grid of MapCells, with some set of MapCells laid on top as pieces.
 */
public class Map extends JComponent implements Cloneable,Serializable,MouseInputListener{
	protected int width,height;
	protected MapCell cells[][];
	protected Vector<Vector<MapCell>> pieces;
	protected Vector<Rectangle> dirty; // TODO
	protected BufferedImage buffer; // TODO
	protected Dimension cellsize;
	protected MapCell currentCell; // The piece or space currently being moused over, if any
	protected Rectangle currentRect; // the bounding rectangle of the current cell
	private Map() {}

	/**
	 * Makes a new map with the specified dimensions, of 32x32 tiles.
	 * @param width
	 * @param height
	 */
	public Map(int width,int height) {
		this(width,height,new MapCell(),new Dimension(32,32));
	}

	/**
	 * Makes a new map with the specified dimensions and cell size.
	 * @param width
	 * @param height
	 * @param cellsize
	 */
	public Map(int width,int height,Dimension cellsize) {
		this(width,height,new MapCell(),cellsize);
	}

	/**
	 * Makes a new map filled with clones of the specified cell.
	 * @param width
	 * @param height
	 * @param cell
	 * @param cellsize
	 */
	public Map(int width,int height,MapCell cell,Dimension cellsize) {
		initMembers(width, height, cellsize);

		layoutComponent(cell);
	}

	protected void initMembers(int width, int height, Dimension cellsize) {
		this.width=width;
		this.height=height;
		this.cellsize=cellsize;
		cells=new MapCell[width][height];
		pieces=new Vector<Vector<MapCell>>(width*height);

		for(int n=0;n!=width*height;n++)
			pieces.add(new Vector<MapCell>());

		buffer=new BufferedImage(getPreferredSize().width,getPreferredSize().height,BufferedImage.TYPE_INT_ARGB);
		dirty=new Vector<Rectangle>();
		dirty.add(new Rectangle(0,0,buffer.getWidth(),buffer.getHeight()));

		addMouseListener(this);
		addMouseMotionListener(this);
	}

	protected void layoutComponent(MapCell cell) {
		cell=(MapCell) cell.clone();

		for(int y=0;y!=height;y++)
			for(int x=0;x!=width;x++)
			{
				cells[x][y]=(MapCell) cell.clone();
				cells[x][y].setConstraints(new MapLayoutConstraints(x,y,0,0));
				cells[x][y].setSize(cellsize);
				cells[x][y].setMap(this);
			}
	}

	protected MapCell[][] getCells(){
		return cells;
	}

	/**
	 * Gets the MapCell on the map layer at the specified coordinates.
	 * @param x
	 * @param y
	 * @return
	 */
	public MapCell getSpace(int x,int y) {
		if(!inBounds(x, y))return null;
		return cells[x][y];
	}

	/**
	 * Puts a MapCell at the coordinates on the map layer.
	 * @param x
	 * @param y
	 * @param cell
	 */
	public void putSpace(int x,int y,MapCell cell) {
		if(!inBounds(x, y))return;
		cells[x][y]=(MapCell) cell.clone();
		cells[x][y].setSize(cellsize);
		cells[x][y].setConstraints(new MapLayoutConstraints(x,y,0,0));
		cells[x][y].setMap(this);
	}



	@Override
	protected void paintComponent(Graphics g) {
		int ix=0,iy=0;
		Graphics2D g2=(Graphics2D) g;
		if(getBorder()!=null){
			Border b=getBorder();
			Insets i=b.getBorderInsets(this);
			ix=i.left;
			iy=i.top;
		}

		g2.translate(ix,iy);

		for(int x=0;x<cells.length;x++){
			for(int y=0;y<cells[x].length;y++){
				Point p=getOrigin(cells[x][y].constraints);
				g2.drawImage(cells[x][y].getCanvas(), p.x,p.y,null);
			}
		}

		for(Vector<MapCell> space : pieces){
			for(MapCell piece : space){
				Point p=getOrigin(piece.constraints);
				g2.drawImage(piece.getCanvas(), p.x,p.y,null);				
			}
		}

		g2.translate(-ix,-iy);
	}

	/**
	 * Returns whether the given point is within the bounds of the map.
	 * @param x
	 * @param y
	 * @return
	 */
	public boolean inBounds(int x, int y) {
		return !(x<0 || y<0 || x>=getMapWidth() || y>=getMapHeight());
	}

	public int getMapHeight() {
		return height;
	}

	public int getMapWidth() {
		return width;
	}

	public Dimension getCellsize() {
		return cellsize;
	}

	public void setCellsize(Dimension cellsize) {
		this.cellsize = cellsize;

		for(int x=0;x!=width;x++)
			for(int y=0;y!=height;y++)
				cells[x][y].setSize(cellsize);

		for(Vector<MapCell> space : pieces)
			if(space!=null && !space.isEmpty())
				for(MapCell piece : space)
					piece.setSize(cellsize);
		revalidate();
		if(getParent() instanceof Container)
			getParent().validate();
	}

	public void putPiece(MapLayoutConstraints constraints,MapCell piece) {
		synchronized(pieces)
		{
			piece.setSize(cellsize);
			piece.setConstraints(constraints);
			piece.setMap(this);
			getPieces(constraints.point.x,constraints.point.y).add(piece);
		}
	}

	public void putPiece(int x,int y,MapCell piece) {
		MapLayoutConstraints cons=new MapLayoutConstraints(x,y);
		putPiece(cons,piece);
	}

	public Vector<MapCell> getPieces(int x,int y){
		synchronized(pieces)
		{
			return pieces.get(x+y*getMapWidth());
		}
	}

	public MapCell getPiece(int x,int y){
		Vector<MapCell> pieces=getPieces(x,y);
		if(pieces.isEmpty())return null;
		return pieces.get(0);
	}

	public void movePiece(MapCell piece,int sx,int sy,int x,int y){
		synchronized(pieces)
		{
			if(getPieces(sx,sy).remove(piece))
				getPieces(x,y).add(piece);

		}
	}

	public void removePiece(MapCell piece){
		Vector<MapCell> forCell=
			getPieces(piece.getConstraints().point.x,
					piece.getConstraints().point.y);
		forCell.remove(piece);
	}


	public Point getOrigin(MapLayoutConstraints loc){
		return new Point(loc.point.x*getCellsize().width+loc.offset.x,
				loc.point.y*getCellsize().height+loc.offset.y);
	}

//	public void paintComponent(Graphics g){		
////	Graphics local;

////	synchronized(pieces)
////	{
////	for(int y=0;y!=getMapHeight();y++)
////	for(int x=0;x!=getMapWidth();x++)
////	{
////	local=g.create(x*cellsize.width,y*cellsize.height,cellsize.width,cellsize.height);
////	local.drawImage(getSpace(x,y).getCanvas(),0,0,null);
////	for(MapCell piece : getPieces(x,y))
////	{
////	int xc=piece.getConstraints().point.x*getCellsize().width+piece.getConstraints().offset.x;
////	int yc=piece.getConstraints().point.y*getCellsize().height+piece.getConstraints().offset.y;
////	local=g.create(x,y,getCellsize().width,getCellsize().height);
////	local.drawImage(piece.getCanvas(),0,0,null);
////	}
////	}
////	}
//	}

	@Override
	public Dimension getMaximumSize() {
		return getPreferredSize();
	}

	@Override
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	@Override
	public Dimension getPreferredSize() {
		int bx=0,by=0;
		if(getBorder()!=null){
			Insets i=getBorder().getBorderInsets(this);
			bx=i.left+i.right;
			by=i.top+i.bottom;
		}
		return new Dimension(bx+getMapWidth()*getCellsize().width,
				by+getMapHeight()*getCellsize().height);
	}

	public void mouseClicked(MouseEvent event) {
		if(currentCell!=null)
			currentCell.mouseClicked(event.getButton());
	}

	public void mousePressed(MouseEvent event) {
		if(currentCell!=null)
			currentCell.mousePressed(event.getButton());		
	}

	public void mouseReleased(MouseEvent event) {
		if(currentCell!=null)
			currentCell.mouseReleased(event.getButton());		
	}

	public void mouseEntered(MouseEvent event) {
		currentCell=findCurrentCell(event.getX(),event.getY());
		if(currentCell!=null)
		{
			currentRect=currentCell.getConstraints().getBounds(getCellsize());
			currentCell.mouseEntered(event.getButton());
		}
	}

	public void mouseExited(MouseEvent event) {
		if(currentCell!=null)
			currentCell.mouseExited(event.getButton());
		currentRect=null;
		currentCell=null;
	}

	public void mouseDragged(MouseEvent event) {
		if(currentCell!=null)
			currentCell.mouseDragged(event.getButton());
	}

	public void mouseMoved(MouseEvent event) {
		MapCell nowOver=findCurrentCell(event.getX(),event.getY());
		if(nowOver!=currentCell)
		{
			if(currentCell!=null)
				currentCell.mouseExited(event.getButton());

			if(nowOver!=null)
				nowOver.mouseEntered(event.getButton());

			currentCell=nowOver;
		}
		else if(currentCell!=null)
			currentCell.mouseMoved(event.getButton());
	}

	/**
	 * @param x
	 * @param y
	 * @return The MapCell that the given coords fall in.
	 */
	protected MapCell findCurrentCell(int x,int y){
		if(!inBounds(x,y))return null;
		MapCell cell=getCells()[x/getCellsize().width][y/getCellsize().height];

//		for(Vector<MapCell> space : pieces)
//		for(MapCell current : space)
//		if(current.getConstraints().getBounds(getCellsize()).contains(x,y))
//		cell=current;

		return cell;
	}

	/**
	 * @param cell A MapCell which is not already in the map
	 * @param start The 
	 * @param end
	 * @param duration
	 * @param steps
	 * @return
	 */
	public MoveThread animateMove(MapCell cell,MapLayoutConstraints end,long duration,int steps) {
		return new MoveThread(cell,cell.getConstraints(),end,duration,steps);
	}

	public MoveThread animateMove(MapCell cell,Point start,Point end,long duration,int steps) {
		return new MoveThread(cell,new MapLayoutConstraints(start),
				new MapLayoutConstraints(end),duration,steps);
	}

	public MoveThread animateMove(MapCell cell,int x1,int y1,int x2,int y2,long duration,int steps) {
		return animateMove(cell,new Point(x1,y1),new Point(x2,y2),duration,steps);
	}

	public MoveThread animateMove(MapCell cell,int x2,int y2,long duration,int steps) {
		return animateMove(cell,
				new MapLayoutConstraints(x2,y2),duration,steps);
	}

	public class MoveThread extends Thread{
		MapLayoutConstraints start,end;
		long duration;
		int steps;
		MapCell cell;

		MoveThread(MapCell cell,MapLayoutConstraints start,MapLayoutConstraints end,long duration,int steps){
			this.start=start;
			this.end=end;
			this.cell=cell;
			this.duration=duration;
			this.steps=steps;
		}

		public void run() {
			Point endCoords=getOrigin(end);
			Point startCoords=getOrigin(start);

			int startX,startY,endX,endY;
			startX=start.point.x;
			startY=start.point.y;
			endX=end.point.x;
			endY=end.point.y;

			double cx=start.offset.x;
			double cy=start.offset.y;
			double dx=(double)(endCoords.x-startCoords.x)/(double)(steps);
			double dy=(double)(endCoords.y-startCoords.y)/(double)(steps);

			MapLayoutConstraints cons=cell.getConstraints();

			try {				
				for(int n=0;n!=steps;n++)
				{
					cx+=dx;
					cy+=dy;
					dirty.add(new Rectangle(getOrigin(cons),getCellsize()));
					cons.offset.x=(int) cx;
					cons.offset.y=(int) cy;
					dirty.add(new Rectangle(getOrigin(cons),getCellsize()));
					repaint();
					sleep(duration/steps);
				}

			}catch(InterruptedException e){}
			finally{
				cell.setConstraints(end);
				movePiece(cell,startX,startY,endX,endY);
				dirty.add(new Rectangle(getOrigin(cons),getCellsize()));
				repaint();
			}
		}
	}

}
